% -----------------------------------------------------------------------------
%
% Anbindung des RDBMS MySQL an Eclipse
% 
% Dateiname:	mysql.pl			Version: 0.0.4
% Autor:	Thomas H. Kolbe			Letzte Aenderung: 27. 9. 2000
%
% Die Einbindung in beliebige Programme kann durch folgende Zeile geschehen:
%
% :- use_module(library(mysql)).
%
% Dazu muss allerdings das Verzeichnis, in welchem sich die zwei Dateien
% mysql.so und mysql.pl befinden, in der Liste der Bibliotheksverzeichnisse
% enthalten sein. Diese Liste kann mittels get_flag(library_path,P) abgefragt,
% dann entsprechend erweitert und durch set_flag(library_path,<neue_Pfadliste>)
% neu gesetzt werden.
% 
% -----------------------------------------------------------------------------

:- module(mysql).

:- comment(include, mysql_comments).

% Es werden alle zur Verfuegung gestellten Praedikate global sichtbar gemacht

:- export
	sql_login/3,
	sql_login/4,
	sql_logout/0,
	sql_insert/3,
	sql_delete/5,
	sql_exec/4,
	sql_select/2,
	sql_query/5,
	sql_querydistinct/5,
	sql_lastmsg/3,
	sql_commit/0,
	sql_rollback/0,
	sql_external/2,
	sql_showdb/0,
	sql_desc/1.


:-	load("mysql.so"),
	external(sql_login/4,"p_login"),
	external(sql_logout/0,"p_logout"),
	external(sql_lastmsg/3,"p_lastmsg"),
	external(sql_exec/1,"p_execsql"),
	external(showtables/0,"p_showtables"),
	external(queryopencursor/7,"p_queryopencursor"),
	external(selectopencursor/3,"p_selectopencursor"),
	external(describeopencursor/3,"p_describeopencursor"),
	b_external(selectfetch/2,"p_selectfetch").

sql_login(Database,User,Pass):-
	sql_login(Database,User,Pass,'localhost').

% Der Select-Befehl ist wegen der Funktionsweise der Eclipse-Funktion
% 'schedule_cut_fail_action(...)' in zwei Teile zerlegt worden.
% (weitere Kommentare in Datei 'oracle.pc')

sql_select(Statement,Varlist):-
	selectopencursor(CursorNr,Statement,Varlist),
	selectfetch(CursorNr,Varlist).

sql_query(Attrlist,Tabellen,Where_Teil,Further_Options,Varlist) :-
	length(Attrlist, N),
	length(Varlist, N),
	queryopencursor(CursorNr,0,Attrlist,Tabellen,Where_Teil,
			Further_Options,Varlist),
	selectfetch(CursorNr,Varlist).

sql_querydistinct(Attrlist,Tabellen,Where_Teil,Further_Options,Varlist) :-
	length(Attrlist, N),
	length(Varlist, N),
	queryopencursor(CursorNr,1,Attrlist,Tabellen,Where_Teil,
			Further_Options,Varlist),
	selectfetch(CursorNr,Varlist).

% sql_showdb/0 gibt alle vom aktuell eingeloggten Benutzer definierten
% Relationen, Sichten und Synonyme aus. Dazu wird der Inhalt
% der von ORACLE definierten Relation 'CAT' ausgelesen und entsprechend
% aufbereitet. 

sql_showdb :-
	nl,
	write('  Table name'),nl,
	write('  ------------------------------'),nl,
	showtables,
	nl.

% sql_desc/1 gibt das Schema einer Relation oder Sicht aus. Die Ausgabe
% erfolgt dabei in aehnlicher Form wie beim SQL*Plus-Befehl 'desc'.
% Der Parameter muss entweder ein String oder ein Atom sein; es wird
% nicht zwischen Gross- und Kleinschreibung unterschieden.  

describe(Tablename,[Field,Type,Null,Key,Default,Extra]) :-
	describeopencursor(CursorNr,Tablename,
	                   [Field,Type,Null,Key,Default,Extra]),
	selectfetch(CursorNr,[Field,Type,Null,Key,Default,Extra]).

sql_desc(Tablename) :-
	nl,
	write('  Name                           Null?    Type'),nl,
	write('  ------------------------------ -------- -------------'),nl,
	describe(Tablename,[Col,Type,Null,_,_,_]),
	printf("  %-30s ",[Col]),
	(Null=='YES' -> write('         ') ;
	                write('NOT NULL ')),
	write(Type),
	nl,
	fail.
sql_desc(_).

% sql_exec/4 fuehrt das im ersten Parameter uebergebene SQL-Kommando
% unmittelbar aus und unifiziert anschliessend die letzten drei Parameter
% mit dem Status der Operationsdurchfuehrung. 

sql_exec(Statement,RC,Msg,Rows) :-
	sql_exec(Statement),
	sql_lastmsg(RC,Msg,Rows).

% MySQL has no transaction management (yet), therefore commit and rollback 
% have no effect. Both predicates sql_commit/0 and sql_rollback/0 are
% always true.

sql_commit :-
	true.

sql_rollback :-
	write("Warning: sql_rollback/0 was called, "),
	write("but MySQL cannot rollback!"),
	nl.

% ins_concat/2 wandelt eine im ersten Parameter uebergebene gemischte 
% Liste aus Strings und Atomen in einen String, in dem die einzelnen
% Elemente durch Kommata voneinander getrennt sind.

ins_concat(AttrList,Result):-
	ins_concat(AttrList,"",Result).
ins_concat([],End,End).
ins_concat([Attr],Tmp,End) :- !,
	(atom(Attr) -> atom_string(Attr,B) ; B=Attr),
	concat_string([Tmp,B],End).
ins_concat([Attr|[A1|As]],Tmp,End) :-
	(atom(Attr) -> atom_string(Attr,B) ; B=Attr),
	concat_string([Tmp,B,","],Tmp2),
	ins_concat([A1|As],Tmp2,End).

% insval_concat/2 wandelt eine im ersten Parameter uebergebene gemischte 
% Liste aus Strings, Atomen, Integer-Zahlen und Real-Zahlen in einen String,
% in dem die Elemente durch Kommata voneinander getrennt sind. Atome und
% Strings werden im Ergebnis-String von Hochkommata (') eingeschlossen. 

insval_concat(ValueList,Result):-
	insval_concat(ValueList,"",Result).
insval_concat([],End,End).
insval_concat([Value],Tmp,End) :- !,
	(atom(Value) -> atom_string(Value,C1), concat_string(["'",C1,"'"],C) ; 
         (number(Value) -> number_string(Value,C) ; 
	  concat_string(["'",Value,"'"],C))),
	concat_string([Tmp,C],End).
insval_concat([Value|[V1|Vs]],Tmp,End) :-
	(atom(Value) -> atom_string(Value,C1), concat_string(["'",C1,"'"],C) ; 
         (number(Value) -> number_string(Value,C) ; 
	  concat_string(["'",Value,"'"],C))),
	concat_string([Tmp,C,","],Tmp2),
	insval_concat([V1|Vs],Tmp2,End).

% sql_insert/3 fuegt ein Tupel in die im ersten Parameter angegebene Relation
% (oder Sicht) ein. Dazu werden im zweiten Parameter eine Liste der Attribute
% und im dritten Parameter die jeweiligen Werte angegeben; die Listen muessen
% also gleichlang sein. Das Praedikat ist wahr, wenn das Tupel erfolgreich
% eingefuegt wurde, ansonsten falsch.  
%
% Beispiel:
%	sql_insert(tabelle1,[id,nr],[1,113]).
%   wird zu folgendem SQL-Befehl:
%	insert into tabelle1(id,nr) values (1,113)

sql_insert(Table,AttrList,ValueList) :-
	length(AttrList,N), 
	length(ValueList,N),
	(atom(Table) -> atom_string(Table,Table2) ; Table2=Table),
	ins_concat(AttrList,Attrs),
	insval_concat(ValueList,Values),
	concat_atom(['insert into ',Table2,'(',Attrs,') values (',Values,')'],
		    Statement),
	sql_exec(Statement,0,_,_).

% del_concat/3 wandelt die in den beiden ersten Parametern uebergebenen Listen
% in einen String um. Die erste Liste kann aus Atomen und Strings zusammen-
% gesetzt sein, die zweite Liste zusaetzlich noch aus Integer- und Real-Zahlen.
% Es wird jeweils ein Element der ersten Liste mit einem Gleichheitszeichen
% und einem Element der zweiten Liste zusammengesetzt; diese Ausdruecke
% werden voneinander durch das Wort 'and' getrennt. Atome und Strings der 
% zweiten Liste werden im Ergebnis-String von Hochkommata (') eingeschlossen.

del_concat(AttrList,ValueList,Result) :-
	del_concat(AttrList,ValueList,"",Result).
del_concat([],[],End,End).
del_concat([Attr],[Value],Tmp,End) :- !,
	(atom(Attr) -> atom_string(Attr,B) ; B=Attr),
	(atom(Value) -> atom_string(Value,C1), concat_string(["'",C1,"'"],C) ; 
         (number(Value) -> number_string(Value,C) ; 
	  concat_string(["'",Value,"'"],C))),
	concat_string([Tmp,B,"=",C],End).
del_concat([Attr|[A1|As]],[Value|[V1|Vs]],Tmp,End) :-
	(atom(Attr) -> atom_string(Attr,B) ; B=Attr),
	(atom(Value) -> atom_string(Value,C1), concat_string(["'",C1,"'"],C) ; 
         (number(Value) -> number_string(Value,C) ; 
	  concat_string(["'",Value,"'"],C))),
	concat_string([Tmp,B,"=",C," and "],Tmp2),
	del_concat([A1|As],[V1|Vs],Tmp2,End).

% sql_delete/5 loescht Tupel aus der im ersten Parameter angegebenen Relation
% (oder Sicht). Die Menge der zu loeschenden Tupel kann sowohl durch die
% im zweiten Parameter anzugebenden 'Where'-Klausel als auch durch die im
% dritten und vierten Parameter anzugebenden Listen eingeschraenkt werden.
% Dabei werden in der ersten Liste die Attribute und in der zweiten Liste
% die jeweiligen Werte der zu loeschenden Tupel angegeben; die Listen muessen
% also gleichlang sein. Der zweite Parameter wird schliesslich noch als weitere
% Bedingung mittels Konjunktion (in SQL: 'and') an den Where-Teil des 
% generierten SQL-Befehls angehaengt. Achtung: Werden weder im zweiten 
% Parameter eine Bedingung (leerer String oder leeres Atom) noch Attribute mit 
% Wertangaben im dritten und vierten Parameter angegeben, so werden alle Tupel
% der bezeichneten Entitaet geloescht! 
%
% Beispiel: 
% 	sql_delete(tabelle1,"id in (select id from tabelle2)",[nr],[17],R).
%   wird zu folgendem SQL-Befehl:
%	delete from tabelle1 where nr=17 and (id in (select id from tabelle2))
%
% Das Praedikat ist wahr, wenn mindestens ein Tupel geloescht wurde und kein
% Datenbankfehler auftrat. Der letzte Parameter wird mit der Anzahl der
% geloeschten Tupel unifiziert.

sql_delete(Table,"",[],[],Rows_Deleted) :-
	!,
	sql_delete(Table,"1=1",[],[],Rows_Deleted).
sql_delete(Table,'',[],[],Rows_Deleted) :-
	!,
	sql_delete(Table,"1=1",[],[],Rows_Deleted).
sql_delete(Table,Where,AttrList,ValueList,Rows_Deleted) :-
	length(AttrList,N), 
	length(ValueList,N),
	(atom(Table) -> atom_string(Table,Table2) ; Table2=Table),
	(atom(Where) -> atom_string(Where,Where2) ; Where2=Where),
	del_concat(AttrList,ValueList,W),
	concat_strings("delete from ",Table2,Statement),
	sql_delete2(Statement,Where2,W,Rows_Deleted).

sql_delete2(Statement,"","",Rows_Deleted) :- !,
	sql_exec(Statement,0,_,Rows_Deleted).
sql_delete2(Statement,Where1,Where2,Rows_Deleted) :-
	concat_string([Statement," where ",Where2],S1),
	(Where2\=="" -> (Where1\=="" -> concat_string([S1," and (",Where1,")"],
						     Statement2) ;
					Statement2=S1) ;
	 concat_strings(S1,Where1,Statement2)),
	sql_exec(Statement2,0,_,Rows_Deleted).

% erzeuge_external/2 generiert ein Praedikat mit einem Funktor, der dem 
% ersten Parameter entspricht, und einer Stelligkeit, die der Laenge der
% im zweiten Parameter angegebenen Liste entspricht. Das erzeugte Praedikat
% bildet eine Ein-Tupel-Schnittstelle zur im ersten Parameter spezifizierten
% Relation oder Sicht mit den im zweiten Parameter aufgelisteten Attributen.
% 
% Beispiel:
%	erzeuge_external(tabelle2,[id,nr]).
%   generiert folgendes Praedikat:
%	tabelle2(_1,_2) :- sql_query([id,nr],tabelle2,'',[_1,_2]).

erzeuge_external(Atom, L) :-
	length(L, N),
	concat_atoms(r_,Atom,Atom2),
	functor(H, Atom2, N),
	H =.. [_| Args],
	B = sql_query(L, Atom, '', '', Args),
	(is_predicate(Atom2/N) -> abolish Atom2/N ; true),
	compile_term(H:-B),
	global Atom2/N.

% sql_external/2 dient zum Einrichten einer Ein-Tupel-Schnittstelle zu einer
% Relation oder Sicht. Der erste Parameter gibt den Namen der Relation an und
% ist gleichzeitig Teil des Funktors des zu generierenden Praedikats; aus 
% diesem Grund muss der Parameter als Atom angegeben sein. 
% Der Funktor setzt sich aus dem Praefix 'r_' und dem ersten Parameter
% zusammen. Der zweite Parameter muss entweder das Atom 'cache' 
% oder 'nocache' sein. Im ersten Fall wird das Praedikat 
% mittels des oben definierten Praedikats 'erzeuge_external/2'
% angelegt (siehe dort). Im zweiten Fall wird hingegen der Inhalt der
% spezifizierten Entitaet vollstaendig eingelesen und als Menge von Atomen
% hinzucompiliert. Bei dieser Variante ist zu beachten, dass sie 
% Datenmanipulationen der Entitaet nicht beruecksichtigt und zu Inkonsistenzen
% fuehren kann. Der Anwender hat selbst darauf zu achten, dass nach
% Manipulationen des Datenbestands der Entitaet diese entweder wieder
% neu "gecached" oder mittels 'sql_external(<Entitaet>,nocache)' als 
% ungepuffert definiert wird.
% Achtung: Beide Varianten beruecksichtigen keine Schemamanipulationen von
% Entitaeten, d.h., nach Aenderung eines Schemas muss die entsprechende
% Ein-Tupel-Schnittstelle neu generiert werden! Dabei ist zu beachten, dass
% wenn sich die Anzahl der Attribute geaendert hat, nach der Generierung
% des neuen Praedikats das alte immer noch existiert, da es ja eine andere
% Stelligkeit als das neue hat und somit durch die Compilierung nicht ersetzt
% wurde. Gegebenenfalls sollte das alte Praedikat dann mittels 'abolish/1'
% entfernt werden.

sql_external(Tablename,nocache) :-
	findall(Col, describe(Tablename,[Col,_,_,_,_,_]),
		Attributes),
	erzeuge_external(Tablename,Attributes).
sql_external(Tablename,cache) :-
	findall(Col, describe(Tablename,[Col,_,_,_,_,_]),
		Attributes),
	length(Attributes,N),
	concat_atoms(r_,Tablename,Tablename2),
	functor(H,Tablename2,N),
	H =.. [_| Args],
	findall(Atom,(sql_query(Attributes,Tablename,'','',Args),
		      Atom =.. [Tablename2 | Args]),All_Atoms),
	(is_predicate(Tablename2/N) -> abolish Tablename2/N ; true),
	findall('',member(_,Attributes),DummyArgs),
	Dummy =.. [Tablename2 | DummyArgs],
	compile_term([Dummy:-fail|All_Atoms]),
	global Tablename2/N.

